import inspect

import jax
import jaxlib

from jaxili.model import ConditionalMAF, MixtureDensityNetwork, ConditionalRealNVP
from jaxili.loss import loss_nll_nle, loss_nll_npe

"""
This script contains static dictionaries that map function names to their respective functions.
Those dictionaries are used to load neural network models using Callable as hyperparameters (e.g. activation functions)
from a JSON file where the name of the function is stored.
"""

#Define a dictionary containing jax activation functions.
def is_jax_nn_activation_function(obj):
    """
    Returns if an object is an activation function from jax.nn.
    """
    return isinstance(obj, jaxlib.xla_extension.PjitFunction) or isinstance(obj, jax._src.custom_derivatives.custom_jvp) or inspect.isfunction(obj)

jax_nn_dict = {
    name: func for name, func in inspect.getmembers(jax.nn, is_jax_nn_activation_function)
}

jax_nn_dict.pop('__getattr__')

#Define a dictionary containing Neural Network from jaxili

jaxili_nn_dict = {
    "ConditionalMAF": ConditionalMAF,
    "MixtureDensityNetwork": MixtureDensityNetwork,
    "ConditionalRealNVP": ConditionalRealNVP,
}

#Define a dictionary containing the loss functions introduced in jaxili
jaxili_loss_dict = {
    'loss_nll_npe': loss_nll_npe,
    'loss_nll_nle': loss_nll_nle,
}


import jax
import jax.numpy as jnp
import flax.linen as nn
import tensorflow_probability as tfp
from typing import Any, Callable

tfp = tfp.experimental.substrates.jax
tfb = tfp.bijectors
tfd = tfp.distributions

class MLPCompressor(nn.Module):
    """
    MLP Compressor

    Defines a MLP compressor to send the summary statistic to the same dimension than the parameters
    """

    hidden_size: list
    activation: Callable
    output_size: int

    @nn.compact
    def __call__(self, x):
        for size in self.hidden_size:
            x = nn.Dense(size)(x)
            x = self.activation(x)
        x = nn.Dense(self.output_size)(x)
        return x


import os
import json
import time
from copy import copy, deepcopy
from typing import Dict, Any, Optional, Callable, Tuple, Iterator, Sequence
from collections import defaultdict
from tqdm import tqdm
import jax
import jax.numpy as jnp
from flax import linen as nn
from flax.training import train_state, checkpoints, orbax_utils
from flax.training.early_stopping import EarlyStopping
import optax
import orbax.checkpoint as ocp

from pytorch_lightning.loggers import TensorBoardLogger, WandbLogger

import jaxili
from jaxili.inventory.func_dict import jax_nn_dict, jaxili_nn_dict, jaxili_loss_dict

class TrainState(train_state.TrainState):
    """
    A simple extension of TrainState to also include batch statistics
    If a model has no batch statistics, it is None.
    Keep an rng state for dropout or init.
    """
    batch_stats : Any = None,
    rng : Any = None

class TrainerModule:

    def __init__(self, 
                 model_class : jaxili.model.NDENetwork,
                 model_hparams : Dict[str, Any],
                 optimizer_hparams : Dict[str, Any],
                 loss_fn : Callable,
                 exmp_input : Any,
                 seed : int = 42,
                 logger_params : Dict[str, Any] = None,
                 enable_progress_bar : bool = True,
                 debug : bool = False,
                 check_val_every_epoch : int = 1,
                 nde_class : str = "NPE",
                 **kwargs):
        """
        A basic Trainer module summarizing most training functionalities
        like logging, model initialization, training loop, etc...

        Atributes
        ---------
        model_class : The class of the model that should be trained.
        model_hparams : A dictionnary of the hyperparameters of the model.
        Is used as input to the model when it is created.
        optimizer_hparams : A dictionnary of the hyperparameters of the optimizer.
        Used during initialization of the optimizer.
        exmp_input : Input to the model for initialisation and tabulate.
        seed : Seed to initialise PRNG.
        logger_params : A dictionary containing the specifications of the logger.
        enable_progress_bar : Whether to enable the progress bar.
        debug : If True, no jitting is applied. Can be helpful for debugging.
        check_val_every_epoch : How often to check the validation set.
        """
        super().__init__()
        self.model_class = model_class
        self.model_hparams = model_hparams
        self.loss_fn = loss_fn
        self.optimizer_hparams = optimizer_hparams
        self.enable_progress_bar = enable_progress_bar
        self.debug = debug
        self.seed = seed
        self.check_val_every_epoch = check_val_every_epoch
        self.nde_class = nde_class
        assert (nde_class=="NPE" or nde_class=="NLE"), ("Choose a valid class of Neural Density Estimator. (NPE or NLE)")
        self.exmp_input = exmp_input
        if self.nde_class=="NLE":
            self.exmp_input = (self.exmp_input[1], self.exmp_input[0])
        self.generate_config(logger_params)
        self.config.update(kwargs)
        #Create an empty model. Note: no parameters yet
        self.model = self.model_class(**self.model_hparams)
        self.init_apply_fn()
        self.print_tabulate(self.exmp_input)
        #Init trainer parts
        self.init_logger(logger_params)
        self.create_jitted_functions()
        self.init_model(self.exmp_input)
        #Initialize checkpointer
        options = ocp.CheckpointManagerOptions(max_to_keep=1, create=True)
        orbax_checkpointer = ocp.PyTreeCheckpointer()
        self.checkpoint_manager = ocp.CheckpointManager(
            self.log_dir, orbax_checkpointer, options
        )

    def init_logger(self,
                    logger_params : Optional[Dict]=None):
        """
        Initialize the logger and created a logging directory.

        Parameters
        ----------
        logger_params : A dictionary containing the specifications of the logger.
        """

        if logger_params is None:
            logger_params = dict()
        #Determine logging directory
        log_dir = logger_params.get('log_dir', None)
        if log_dir is None:
            base_log_dir = logger_params.get('base_log_dir', 'checkpoints/')
            #Prepare logging
            log_dir = os.path.join(base_log_dir, self.config['model_class'])
            if 'logger_name' in logger_params:
                log_dir = os.path.join(log_dir, logger_params['logger_name'])
            version = None
        else:
            version = ''
        #Create logger object
        logger_type = logger_params.get('logger_type', 'TensorBoard').lower()
        if logger_type == 'tensorboard':
            self.logger = TensorBoardLogger(save_dir=log_dir, version=version, name='')
        elif logger_type == 'wandb':
            self.logger = WandbLogger(save_dir=log_dir, version=version, name='')
        else:
            assert False, f'Unknown logger type \"{logger_type}\"'
        #Save hyperparameters
        log_dir = self.logger.log_dir
        if not os.path.isfile(os.path.join(log_dir, 'hparams.json')):
            os.makedirs(os.path.join(log_dir, 'metrics/'), exist_ok=True)
            with open(os.path.join(log_dir, 'hparams.json'), 'w') as f:
                json.dump(self.config, f, indent=4)
        self.log_dir = log_dir

    def init_model(self, 
                   exmp_input : Any):
        """
        Creates an initial training state with newly generated network parameters

        Parameters
        ----------
        exmp_input : An input to the model with which the shapes are inferred.
        """
        #Prepare PRNG and input
        model_rng = jax.random.PRNGKey(self.seed)
        model_rng, init_rng = jax.random.split(model_rng)
        exmp_input = [exmp_input] if not isinstance(exmp_input, (list, tuple)) else exmp_input
        #Run model initialization
        variables = self.run_model_init(exmp_input, init_rng)
        #Create default state. Optimizer is initialized later
        self.state = TrainState(step=0,
                                apply_fn=self.apply_fn,
                                params=variables['params'],
                                batch_stats=variables.get('batch_stats'),
                                rng=model_rng,
                                tx=None,
                                opt_state=None)
        
    def init_apply_fn(self):
        """
        Initialize a default apply function for the model.
        """
        self.apply_fn = self.model.log_prob

    def generate_config(self, logger_params):
        """
        Generates a configuration dictionary for the trainer.
        """
        self.config = {
            'model_class': self.model_class.__name__,
            'model_hparams': deepcopy(self.model_hparams),
            'loss_fn': self.loss_fn.__name__,
            'optimizer_hparams': self.optimizer_hparams,
            'logger_params': logger_params,
            'enable_progress_bar': self.enable_progress_bar,
            'debug': self.debug,
            'check_val_every_epoch': self.check_val_every_epoch,
            'seed': self.seed
        }



        if 'activation' in self.model_hparams.keys():
            self.config['model_hparams']['activation'] = self.model_hparams['activation'].__name__
        
    def run_model_init(self,
                       exmp_input : Any,
                       init_rng : Any) -> Dict:
        """
        The model initialization call

        Parameters
        ----------
        exmp_input : An input to the model with which the shapes are inferred.
        init_rng : A jax.random.PRNGKey

        Returns
        -------
            The initialized variable dictionary.
        """
        return self.model.init(init_rng, *exmp_input, method='log_prob')
    
    def print_tabulate(self, 
                       exmp_input : Any):
        """
        Prints a summary of the model represented as a table.

        Parameters
        ----------
        exmp_input : An input to the model with which the shapes are inferred.
        """
        try:
            print(self.model.tabulate(jax.random.PRNGKey(0), *exmp_input))
        except Exception as e:
            print(f'Could not tabulate model: {e}')

    def init_optimizer(self,
                       num_epochs : int,
                       num_steps_per_epoch : int):
        """
        Initializes the optimizer and learning rate scheduler.

        Parameters
        ----------
        num_epochs : Number of epochs to train.
        num_steps_per_epoch : Number of steps per epoch.
        """

        hparams = copy(self.optimizer_hparams)

        #Initialize optimizer
        optimizer_name = hparams.pop('optimizer_name', 'adam')
        if optimizer_name.lower() == 'adam':
            opt_class = optax.adam
        elif optimizer_name.lower() == 'sgd':
            opt_class = optax.sgd
        elif optimizer_name.lower() == 'adamw':
            opt_class = optax.adamw
        else:
            assert False, f'Unknown optimizer \"{optimizer_name}\"'
        #Initialize learning rate scheduler
        #A cosine decay scheduler is used, but others are also possible
        lr = hparams.pop('lr', 1e-3)
        warmup = hparams.pop('warmup', 0.1)
        lr_schedule = optax.warmup_cosine_decay_schedule(
            init_value=0.0,
            peak_value=lr,
            warmup_steps=warmup,
            decay_steps=int(num_epochs * num_steps_per_epoch),
            end_value=0.01 * lr
        )
        #Clip gradients at max value, and evt. apply weight decay
        transf = [optax.clip_by_global_norm(hparams.pop('gradient_clip', 5.0))]
        if opt_class == optax.sgd and 'weight_decay' in hparams:
            transf.append(optax.add_decayed_weights(hparams.pop('weight_decay', 0.0)))
        optimizer = optax.chain(
            *transf,
            opt_class(lr_schedule, **hparams)
        )
        #Initialize training state
        self.state = TrainState.create(
            apply_fn=self.state.apply_fn,
            params=self.state.params,
            batch_stats=self.state.batch_stats,
            tx=optimizer,
            rng=self.state.rng
        )

    def create_jitted_functions(self):
        """
        Creates jitted versions of the training, validation and evaluation functions.
        If self.debug is True, no jitting is applied.
        """

        train_step, eval_step = self.create_functions()
        if self.debug: #Skip jitting
            print('Skipping jitting due to debug=True')
            self.train_step = train_step
            self.eval_step = eval_step
        else:
            self.train_step = jax.jit(train_step)
            self.eval_step = jax.jit(eval_step)

    def create_functions(self) -> Tuple[Callable[[TrainState, Any], Tuple[TrainState, Dict]],
                                       Callable[[TrainState, Any], Tuple[TrainState, Dict]]]:
        """
        Creates and returns functions for the training and evaluation step.
        The functions take as input the training state and a batch from the train/val/test loader.
        Both functions are expected to return a dictionary of logging metrics,
        and the training function a new train state. This function needs to be
        overwritten by a subclass. The train_step and eval_step functions
        here are examples for the signature of the functions.
        """
        def train_step(state: TrainState,
                       batch : Any):
            loss_fn = lambda params: self.loss_fn(self.model, params, batch)
            loss, grads = jax.value_and_grad(loss_fn)(state.params)
            state = state.apply_gradients(grads=grads)
            metrics = {'loss': loss}
            return state, metrics
        
        def eval_step(state : TrainState,
                      batch : Any):
            loss = self.loss_fn(self.model, state.params, batch)
            metrics = {'loss': loss}
            return metrics
        
        return train_step, eval_step
    
    def train_model(self,
                    train_loader : Iterator,
                    val_loader : Iterator,
                    test_loader : Optional[Iterator] = None,
                    num_epochs : int = 500,
                    min_delta : float = 1e-3,
                    patience : int = 20) -> Dict[str, Any]:
        """
        Starts a training loop for the given number of epochs.

        Parameters
        ----------
        train_loader : An iterator over the training data.
        val_loader : An iterator over the validation data.
        test_loader : If given, best model will be evaluated on the test set.
        num_epochs : Number of epochs for which to train the model.
        min_delta : Minimum change in the monitored metric to qualify as an improvement.
        patience : Number of epochs with no improvement after which training will be stopped.

        Returns
        -------
        A dictionary of the train, validation and evt. test metrics for
        the best model on the validation set.
        """

        #Create optimizer and the scheduler for the given numer of epochs
        self.init_optimizer(num_epochs, len(train_loader))
        #Prepare training loop
        self.on_training_start()
        best_eval_metrics = None
        best_epoch = None
        early_stop = EarlyStopping(min_delta, patience)
        pbar = self.tracker(range(1, num_epochs+1), desc='Epochs')
        for epoch_idx in pbar:
            train_metrics = self.train_epoch(train_loader)
            self.logger.log_metrics(train_metrics, step=epoch_idx)
            self.on_training_epoch_end(epoch_idx)
            #Validation every N epochs
            if epoch_idx % self.check_val_every_epoch == 0:
                eval_metrics = self.eval_model(val_loader, log_prefix='val/')
                self.on_validation_epoch_end(epoch_idx, eval_metrics, val_loader)
                self.logger.log_metrics(eval_metrics, step=epoch_idx)
                self.save_metrics(f'eval_epoch_{str(epoch_idx).zfill(3)}', eval_metrics)
                #Save best model
                if self.is_new_model_better(eval_metrics, best_eval_metrics):
                    best_eval_metrics = eval_metrics
                    best_eval_metrics.update(train_metrics)
                    best_epoch = epoch_idx
                    self.save_model(step=epoch_idx)
                    self.save_metrics('best_eval', best_eval_metrics)
                early_stop = early_stop.update(eval_metrics['val/loss'])
                if early_stop.should_stop:
                    print(f'Neural network training stopped after {epoch_idx} epochs.')
                    print(f'Early stopping with best validation metric: {early_stop.best_metric}')
                    print(f'Best model saved at epoch {best_epoch}')
                    print(f'Early stopping parameters: min_delta={min_delta}, patience={patience}')
                    break
                if self.enable_progress_bar:    
                    pbar.set_description(f"Epochs: Val loss {eval_metrics['val/loss']:.3f}/ Best val loss {early_stop.best_metric:.3f}")
        #Test best model if possible
        if test_loader is not None:
            #self.load_model()
            test_metrics = self.eval_model(test_loader, log_prefix='test/')
            self.logger.log_metrics(test_metrics, step=epoch_idx)
            self.save_metrics('test', test_metrics)
            best_eval_metrics.update(test_metrics)
        #Close logger
        self.logger.finalize('success')
        return best_eval_metrics
    
    def train_epoch(self,
                    train_loader : Iterator) -> Dict[str, Any]:
        """
        Trains the model for one epoch

        Parameters
        ----------
        train_loader : An iterator over the training data.

        Returns
        -------
        A dictionary of the average training metrics over all batches for logging
        """

        #Train model for one epoch, and log avg loss and accuracy
        metrics = defaultdict(float)
        num_train_steps = len(train_loader)
        start_time = time.time()
        for batch in train_loader:
            self.state, step_metrics = self.train_step(self.state, batch)
            for key in step_metrics:
                metrics['train/'+key] += step_metrics[key] / num_train_steps
        metrics = {key: metrics[key].item() for key in metrics}
        metrics['epoch_time'] = time.time() - start_time
        return metrics

    def eval_model(self,
                   data_loader : Iterator,
                   log_prefix : Optional[str] = '') -> Dict[str, Any]:
        """
        Evaluates the model on a dataset.

        Parameters
        ----------
        data_loader : An iterator over the data.
        log_prefix : A prefix to add to all metrics.

        Returns
        -------
        A dictionary of the evaluation metrics, averaged over data points in the dataset
        """

        #Test model on all element of the dataloader and return avg loss
        metrics = defaultdict(float)
        num_elements = 0
        for batch in data_loader:
            step_metrics = self.eval_step(self.state, batch)
            batch_size = batch[0].shape[0] if isinstance(batch, (list, tuple)) else batch.shape[0]
            for key in step_metrics:
                metrics[key] += step_metrics[key] *batch_size
            num_elements += batch_size
        metrics = {(log_prefix + key): (metrics[key] / num_elements).item() for key in metrics}
        return metrics
    
    def is_new_model_better(self,
                            new_metrics : Dict[str, Any],
                            old_metrics : Dict[str, Any]) -> bool:
        """
        Compares two sets of evaluation metrics to decide whether the new
        model is better than the previous ones or not.

        Parameters
        ----------
        new_metrics : A dictionary of the evaluation metrics of the new model
        old_metrics : A dictionary of the evaluation metrics of the previously best model.

        Returns
        -------
        True if the new model is better, False otherwise.
        """

        if old_metrics is None:
            return True
        for key, is_larger in [('val/val_metric', False), ('val/acc', True), ('val/loss', False)]:
            if key in new_metrics:
                if is_larger:
                    return new_metrics[key] > old_metrics[key]
                else:
                    return new_metrics[key] < old_metrics[key]
        assert False, f'No known metrics to log on: {new_metrics}'

    def tracker(self,
                iterator : Iterator,
                **kwargs) -> Iterator:
        """
        Wraps an iterator in a progress bar track (tqdm) if the progress
        bat is enabled.

        Parameters
        ----------
        iterator : iterator to wrap in tqdm
        kwargs : additional arguments to tqdm

        Returns
        -------
        Wrapped iterator if progress bar is enabled, otherwise same iterator than input
        """

        if self.enable_progress_bar:
            return tqdm(iterator, **kwargs)
        else:
            return iterator
        
    def save_metrics(self,
                     filename : str,
                     metrics : Dict[str, Any]):
        """
        Saves a dictionary of metrics to file. Can be used as a textual
        representation of the validation performance for checking in the terminal

        Parameters
        ----------
        filename : Name of the metrics file without folders and postfix
        metrics : A dictionary of the metrics to save
        """
        with open(os.path.join(self.log_dir, f'metrics/{filename}.json'), 'w') as f:
            json.dump(metrics, f, indent=4)
        
    def on_training_start(self):
        """
        Method called before training is started. Can be used for 
        additional initialization operations etc.
        """
        pass

    def on_training_epoch_end(self,
                              epoch_idx : int):
        """
        Method called at the end of each training epoch. Can be used for
        additional logging or similar.

        Parameters
        ----------
        epoch_idx : Index of the epoch that just finished.
        """
        pass

    def on_validation_epoch_end(self,
                                epoch_idx : int,
                                eval_metrics : Dict[str, Any],
                                val_loader : Iterator):
        """
        Method called at the end of each validation epoch. Can be used for
        additional logging and evaluation.

        Parameters
        ----------
        epoch_idx : Index of the epoch that just finished.
        eval_metrics : A dictionary of the evaluation metrics.
        val_loader : DataLoader of the validation set to support additional evaluation.
        """
        pass

    def save_model(self, 
                   step : int =0):
        """
        Saves current training state at certain training iteration. Only the model
        parameters and batch statistics are saved to reduce memory footprint.
        To support the training to be continued from a checkpoint, this method
        can be extended to include the optimizer state as well.

        Parameters
        ----------
        step : Index of the step to save the model at, e.g. epoch
        """
        target = {'params': self.state.params,
                    'batch_stats': self.state.batch_stats}
        save_args = orbax_utils.save_args_from_target(target)
        self.checkpoint_manager.save(step, target, save_kwargs={'save_args': save_args})
        
    def load_model(self):
        """
        Loads model and batch statistics from the logging directory
        """
        step = self.checkpoint_manager.latest_step()
        state_dict = self.checkpoint_manager.restore(step)
        self.state = TrainState.create(
            apply_fn=self.apply_fn,
            params=state_dict['params'],
            batch_stats=state_dict['batch_stats'],
            tx=self.state.tx if self.state.tx else optax.sgd(0.1),
            rng=self.state.rng
        )
    
    def bind_model(self):
        """
        Returns a model with parameters bound to it. Enables an easier inference access.

        Returns
        -------
        The model with parameters and evt. batch statistics bound to it.
        """
        params = {'params': self.state.params}
        if self.state.batch_stats:
            params['batch_stats'] = self.state.batch_stats
        return self.model.bind(params)
    
    @classmethod
    def load_from_checkpoints(cls,
                              model_class : jaxili.model.NDENetwork,
                              checkpoint : str,
                              exmp_input : Any) -> Any:
        """
        Creates a Trainer object with same hyperparameters and loaded model from
        a checkpoint directory.

        Parameters
        ----------
        checkpoint : Folder in which the checkpoint and hyperparameter file is stored
        exmp_input : An input to the model with which the shapes are inferred.

        Returns
        -------
        A Trainer object with model loaded from the checkpoint folder.
        """

        hparams_file = os.path.join(checkpoint, 'hparams.json')
        assert os.path.isfile(hparams_file), 'Could not find hparams file.'
        with open(hparams_file, 'r') as f:
            hparams = json.load(f)
        assert hparams['model_class'] == model_class.__name__, 'Model class does not match. Check that you are using the correct architecture.'
        hparams.pop('model_class')
        #Check if an activation function is used as a hyperparameter if the neural network.
        assert hparams['loss_fn'] in jaxili_loss_dict, 'Unknown loss function. Check that the loss function you used comes from `jax.nn`.'
        hparams['loss_fn'] = jaxili_loss_dict[hparams['loss_fn']]
        if 'activation' in hparams['model_hparams'].keys():
            hparams['model_hparams']['activation'] = jax_nn_dict[hparams['model_hparams']['activation']]
        if not hparams['logger_params']:
            hparams['logger_params'] = dict()
        hparams['logger_params']['log_dir'] = checkpoint
        trainer = cls(model_class = model_class, exmp_input = exmp_input, **hparams)
        trainer.load_model()
        return trainer



import numpy as np
from typing import Sequence, Union, Any
import jax.numpy as jnp

import torch
import torch.utils.data as data

def numpy_collate(batch):
    if isinstance(batch[0], np.ndarray):
        return np.stack(batch)
    elif isinstance(batch[0], (tuple, list)):
        transposed = zip(*batch)
        return [numpy_collate(samples) for samples in transposed]
    elif isinstance(batch[0], jnp.ndarray):
        return jnp.stack(batch)
    else:
        return np.array(batch)
    
def create_data_loader(*datasets : Sequence[data.Dataset],
                       train : Union[bool, Sequence[bool]]=True,
                       batch_size : int = 128,
                       num_workers : int = 4,
                       seed : int = 42):
    """
    Creates data loaders used in JAX for a set of datasets.

    Parameters
    ----------
    datasets : Datasets for which data loaders are created.
    train : Sequence indicating which datasets are used for training and which not.
    If single bool, the same value is used for all datasets.
    batch_size : Batch size to use in the data loaders.
    num_workers : Number of workers for each datasets.
    seed : Seed to initalize the workers and shuffling with
    """
    loaders = []
    if not isinstance(train, (list, tuple)):
        train = [train for _ in datasets]
    for dataset, is_train in zip(datasets, train):
        loader = data.DataLoader(dataset,
                                 batch_size=batch_size,
                                 shuffle=is_train,
                                 drop_last=is_train,
                                 collate_fn=numpy_collate,
                                 num_workers=num_workers,
                                 persistent_workers=is_train,
                                 generator=torch.Generator().manual_seed(seed))
        loaders.append(loader)
    return loaders

def check_density_estimator(estimator_arg: str):
    """
    Check density estimator argument to see if it belongs to the authorized network.

    Parameters
    ----------
    estimator_arg : str
        Density estimator argument to check.
    """
    if estimator_arg not in ['maf', 'mdn', 'realnvp']:
        raise ValueError(f"Invalid density estimator argument: {estimator_arg}. Options are 'maf', 'mdn' and 'realnvp'.")
    
def validate_theta_x(theta: Any, x: Any):
    """
    Checks if the passed $(\theta, x)$ pair is valid.

    We check that:
    - $\theta$ and $x$ are jax arrays
    - $\theta$ and $x$ have the same number of samples.
    - $\theta$ and $x$ have dtype=float32.

    Raises:
        AssertionError if $\theta$ and $x$ are not jax arrays, do not have the same batch size or are not dtype==np.float32.
    
    Parameters
    ----------
    theta : Any
        Parameters of the simulations.
    x : Any
        Simulation outputs.
    """

    assert isinstance(theta, jnp.ndarray), "theta should be a jax array."
    assert isinstance(x, jnp.ndarray), "x should be a jax array."
    assert theta.shape[0] == x.shape[0], (
        f"Number of parameter sets ({theta.shape[0]}) and number of simulation outputs ({x.shape[0]}) should be the same."
    )

    assert theta.dtype == jnp.float32, "theta should have dtype float32."
    assert x.dtype == jnp.float32, "x should have dtype float32."

    batch_size = theta.shape[0]
    
    return theta, x, batch_size



import jax
import jax.numpy as jnp

from typing import Any
from jaxtyping import Array, Float, PyTree

def loss_nll_npe(model: Any, params: PyTree, batch: Any)->Array:
    """
    Negative log-likelihood loss function for NPE methods
    using a given neural network as a model.
    In NPE, the log-probability is given by the density estimator in
    parameter space conditioned on the data.

    Parameters
    ----------
    model : Any
        Neural network model from `jaxili.model`.
    params : PyTree
        Parameters of the neural network.
    batch : Any
        Batch of (parameters, outputs) to compute the loss.

    Returns
    -------
    Array
        Mean of the negative log-likelihood loss across the batch.
    """

    thetas, xs = batch

    output =  model.apply({'params': params}, thetas, xs, method='log_prob')
    return -jnp.mean(output)

def loss_nll_nle(model: Any, params: PyTree, batch: Any):
    """
    Negative log-likelihood loss function for NLE methods
    using a given neural network as a model.
    In NPE, the log-probability is given by the density estimator in
    parameter space conditioned on the data.

    Parameter
    ---------
    model : Any
        Neural network model from `jaxili.model`.
    params : PyTree
        Parameters of the neural network.
    batch : Any
        Batch of (parameters, outputs) to compute the loss.

    Returns
    -------
    Array
        Mean of the negative log-likelihood loss across the batch.
    """

    thetas, xs = batch

    return -jnp.mean(model.apply({'params': params}, xs, thetas, method='log_prob'))


from flax import linen as nn
import distrax
import numpy as np
import jax
import jax.numpy as jnp
import tensorflow_probability as tfp
from typing import Any, Optional, Callable
from jax.scipy.stats import multivariate_normal
from functools import partial 

tfp = tfp.experimental.substrates.jax
tfb = tfp.bijectors
tfd = tfp.distributions

class NDENetwork(nn.Module):
    """
    A Normalizing Flow parent class to implement normalizing flows using
    neural networks.
    """
    def log_prob(self, x, y=None, **kwargs):
        """
        Log probability of the data point x conditioned by y.
        """
        raise NotImplementedError("log_prob method not implemented in your child class of NDENetwork")

    def sample(self, y, num_samples, key):
        """
        Sample from the distribution conditioned by y.
        """
        raise NotImplementedError("sample method not implemented in tour child class of NDENetwork")

class MixtureDensityNetwork(NDENetwork):
    """
    A Mixture of Gaussian Density modeled using neural networks.
    The weights of each gaussian component, the mean and the covariance are learned by the network.
    """
    n_data : int #Dimension of data vector
    n_components : int #number of mixture components
    layers : list #list of hidden layers size
    activation : Callable #activation function

    @nn.compact
    def __call__(self, y, **kwargs):
        """
        Builds a bijector that tranforms a multivariate Gaussian distribution
        into a Mixture of Gaussian distribution using a neural network.
        The weights, means and covariances are obtained from a conditioned variable y.

        Parameters
        ----------
        y : jnp.Array
            Conditionning variable

        Returns
        -------
        distribution : tfd.Distribution
            Mixture of Gaussian distribution
        """
        kernel_init = kwargs.get('kernel_init', nn.initializers.variance_scaling(scale=1., mode='fan_in', distribution='normal'))
        for size in self.layers:
            y = self.activation(nn.Dense(size, kernel_init=kernel_init)(y))
        final_size = self.n_components * (1 + self.n_data + self.n_data*(self.n_data+1)//2)
        y = nn.Dense(final_size, kernel_init=kernel_init)(y)
        logits = jax.nn.log_softmax(y[..., :self.n_components])
        locs = y[..., self.n_components:self.n_components*(self.n_data+1)]
        scale_tril = y[..., self.n_components*(self.n_data+1):]

        distribution = distrax.MixtureSameFamily(
            mixture_distribution=distrax.Categorical(logits=logits),
            components_distribution=tfd.MultivariateNormalTriL(
                loc=jnp.reshape(locs, (-1, self.n_components, self.n_data)),
                scale_tril=tfp.math.fill_triangular(jnp.reshape(scale_tril, (-1, self.n_components, self.n_data*(self.n_data+1)//2)))
            )
        )

        return distribution

    def log_prob(self, x, y, **kwargs):
        """
        Returns the log probability of the data point x conditioned by y.

        Parameters
        ----------
        x : jnp.Array
            Data point
        y : jnp.Array
            Conditionning variable

        Returns
        -------
        log_prob : jnp.Array
            Log probability of the data point
        """
        distribution = self.__call__(y, **kwargs)
        return distribution.log_prob(x)

    def sample(self, y, num_samples, key):
        """
        Samples from the distribution conditioned by y.

        Parameters
        ----------
        y : jnp.Array
            Conditionning variable
        num_samples : int
            Number of samples
        key : jnp.Array
            Random key

        Returns
        -------
        samples : jnp.Array
            num_samples samples from the distribution
        """
        distribution = self.__call__(y)
        return distribution.sample(sample_shape=num_samples, seed=key).squeeze()

class AffineCoupling(nn.Module):
    y : Any #Conditionning variable
    layers : list #list of hidden layers size
    activation : callable #activation function

    @nn.compact
    def __call__(self, x, output_units, **kwargs):
        x = jnp.concatenate([x, self.y], axis=-1)
        for i, layer_size in enumerate(self.layers):
            x = self.activation(nn.Dense(layer_size, kernel_init=nn.initializers.truncated_normal(0.001))(x))
        
        #Shift and Scale parameters
        shift = nn.Dense(output_units, kernel_init=nn.initializers.truncated_normal(0.001))(x)
        scale = nn.softplus(nn.Dense(output_units, kernel_init=nn.initializers.truncated_normal(0.001))(x)) + 1e-3

        return tfb.Chain([tfb.Shift(shift), tfb.Scale(scale)])
    
class ConditionalRealNVP(NDENetwork):
    n_in : int #Dimension of the input
    n_layers : int #Number of layers
    layers : list[int] #list of hidden layers size
    activation : Callable #activation function

    @nn.compact
    def __call__(self, y, **kwargs):
        """
        Build the bijector using tensorflow_probability

        Parameters
        ----------
        y : jnp.Array
            Conditionning variable

        Returns
        -------
        nvp : tfd.Distributions
            Normalizing Flow transporting a multidimensional Gaussian
            to a more complex distribution.
        """
        bijector_fn  = partial(
            AffineCoupling,
            layers=self.layers,
            activation=self.activation
        )
        base_distribution = distrax.MultivariateNormalDiag(jnp.zeros(self.n_in), jnp.ones(self.n_in))
        chain = distrax.Chain(
            [
                tfb.Permute(jnp.arange(self.n_in)[::-1])(
                    tfb.RealNVP(
                        self.n_in//2, bijector_fn=bijector_fn(y, name="b%d" % i)
                    )
                )
            for i in range(self.n_layers)
            ]
        )

        nvp = distrax.Transformed(
            base_distribution,
            bijector=chain
        )

        return nvp
    
    def sample(self, y, num_samples, key, **kwargs):
        """
        Samples from the distribution mapped by the real NVP

        Parameters
        ----------
        y : jnp.Array
            Conditionning variable
        num_samples : int
            Number of samples
        key : jnp.Array
            Random key

        Returns
        -------
        samples : jnp.Array
            num_samples samples from the distribution
        """
        nvp = self.__call__(y)
        return nvp.sample(sample_shape=num_samples, seed=key)

    def log_prob(self, x, y, **kwargs):
        """
        Computes the log probability of the data point x conditioned by y from the normalizing flow.

        Parameters
        ----------
        x : jnp.Array
            Data point
        y : jnp.Array
            Conditionning variable
        """
        nvp = self.__call__(y)
        return nvp.log_prob(x)


#Reproduce implementation of MADE and MAFs from https://github.com/e-hulten/maf/blob/master/made.py

class MaskedLinear(nn.Module): #Check if there is no issue when you jit a loss using such a network. Note: Mask will not change after initialisation.
    """
    Linear transformation with masked out elements.

    y = x.dot(mask*W.T)+b

    Parameters
    ----------
    n_out : int
        Output dimension
    bias : bool
        Whether to include bias. Default True.
    """
    n_out : int
    bias : bool = True
    mask: Any = None

    def initialize_mask(self, mask : Any):
        """Internal method to initialize mask"""
        self.mask = mask

    @nn.compact
    def __call__(self, x):
        """Apply masked linear transformation"""
        layer = nn.Dense(self.n_out, use_bias=self.bias, param_dtype=jnp.float64, kernel_init=nn.initializers.truncated_normal(0.01))
        is_initialized = self.has_variable('params', 'Dense_0')
        if is_initialized: 
            w = layer.variables['params']['kernel']
            b = layer.variables['params']['bias']
        else:
            return layer(x)
        return jnp.dot(x, self.mask*w)+b

class ConditionalMADE(nn.Module):
    n_in : int #Size of the input vector
    hidden_dims : list[int] #list of hidden dimensions
    activation : Callable #Activation function
    n_cond : int =0 #Size of the conditionning variable. 0 if None.
    gaussian : bool = True #whether the output are mean and variance of a Gaussian conditional
    random_order : bool = False #Whether to use random order of the input for masking
    seed : Optional[int] = None #Random seed to label nodes
    

    def setup(self):
        
        np.random.seed(self.seed)
        self.n_out = 2*self.n_in if self.gaussian else self.n_in
        masks = {}
        mask_matrix = []
        layers = []

        dim_list = [self.n_in+self.n_cond, *self.hidden_dims, self.n_out]

        #Make layers and activation functions
        for i in range(len(dim_list)-2):
            layers.append(MaskedLinear(dim_list[i+1]))
            layers.append(self.activation)
        #Last hidden layer to output layer
        layers.append(MaskedLinear(dim_list[-1]))
        #Create masks
        self._create_masks(mask_matrix, masks, layers)
        #Create model
        self.layers = layers
        self.model = nn.Sequential(self.layers)

    def _create_masks(self, mask_matrix: list, masks: dict, layers : list):
        """Create masks for the model"""
        L = len(self.hidden_dims) #Number of hidden layers
        D = self.n_in #Number of input parameters
        C = self.n_cond #Number of conditionning parameters

        #Whather to use random or natural order of the input
        masks[0] = np.random.permutation(D) if self.random_order else np.arange(D)

        #Set the connectivity number for the hidden layers
        for l in range(L):
            low = masks[l].min() #Get the lowest index in the previous layer
            size = self.hidden_dims[l] #The size of the current hidden layer
            masks[l+1] = np.random.randint(low, D-1, size=size)
        
        #Order of the output layer is the same as the input layer
        masks[L+1] = masks[0]

        #Create mask matric for input -> hidden_layer_1
        m = masks[0]
        m_next = masks[1]
        M = np.ones((len(m), len(m_next)))
        for j in range(len(m_next)):
            M[:, j] = (m <= m_next[j]).astype(int)
        M_cond = np.ones((C, len(m_next)))
        M = np.concatenate([M, M_cond], axis=0)
        mask_matrix.append(jnp.array(M))


        #Create mask matrix for hidden_layer_1 -> ... -> last_hidden_layers
        for i in range(1, len(masks)-2):
            m = masks[i]
            m_next = masks[i+1]
            #Initialise mask matrix
            M = np.zeros((len(m), len(m_next)))
            for j in range(len(m_next)):
                #Compare m_next[j] to each element of m
                M[:, j] = (m <= m_next[j]).astype(int)
            #append matrix to mask list
            mask_matrix.append(jnp.array(M))

        #Create mask matrix for last_hidden_layer -> output
        m = masks[len(masks)-2]
        m_next = masks[len(masks)-1]
        M = np.zeros((len(m), len(m_next)))
        for j in range(len(m)):
            #Compare m_next[j] to each element of m
            M[j, :] = (m[j] < m_next).astype(int)
        #append matrix to mask list
        mask_matrix.append(jnp.array(M))

        #If the output is Gaussian, double the number of output (mu, sigma)
        #Pairwise identical mask
        if self.gaussian:
            m = mask_matrix.pop(-1)
            mask_matrix.append(jnp.concatenate([m, m], axis=1))

        #Initialize the MaskedLinear layers with weights
        mask_iter = iter(mask_matrix)
        for module in layers:
            if isinstance(module, MaskedLinear):
                module.initialize_mask(next(mask_iter))

    def __call__(self, x, y=None):
        """
        Forward pass of the model

        Parameters
        ----------
        x : jnp.Array
            Input vector
        y : jnp.Array
            Conditionning variable
        """
        if self.n_cond != 0:
            x = jnp.concatenate([x, y], axis=-1)
        if self.gaussian:
            return self.model(x)
        else:
            return jax.nn.sigmoid(self.model(x))
        
class MAFLayer(nn.Module):
    n_in : int #Size of the input vector
    n_cond : int #Size of the conditionning variable
    hidden_dims : list[int] #list of hidden dimensions
    reverse : bool #Whether to reverse the order of the input
    activation : Callable #Activation function
    seed : Optional[int] = None #Random seed to label nodes

    def forward(self, x, y=None):
        out = self.__call__(x, y)
        mu, logp = jnp.split(out, 2, axis=-1)
        u = (x-mu)*jnp.exp(0.5*logp)
        u = jnp.flip(u, axis=-1) if self.reverse else u
        log_det = 0.5*jnp.sum(logp, axis=-1)
        return u, log_det
    
    def backward(self, u, y=None):
        u = jnp.flip(u, axis=-1) if self.reverse else u
        x = jnp.zeros_like(u)
        for dim in range(self.n_in):
            out = self.__call__(x,y)
            mu, logp = jnp.split(out, 2, axis=-1)
            mod_logp = jax.lax.clamp(-jnp.inf, -0.5*logp, max=10.)
            x = x.at[:,dim].set(mu[:,dim]+jnp.exp(mod_logp[:,dim])*u[:,dim])
        log_det = jnp.sum(mod_logp, axis=-1)
        return x, log_det
        

    @nn.compact
    def __call__(self, x, y=None):
        """
        Forward pass of the model. Returns mean and variance of the gaussian conditionals.

        Parameters
        ----------
        x : jnp.Array
            Input vector
        y : jnp.Array
            Conditionning variable
        """
        x = ConditionalMADE(n_in=self.n_in, hidden_dims=self.hidden_dims, n_cond=self.n_cond, seed=self.seed, activation=self.activation)(x, y)
        return x
        
class ConditionalMAF(NDENetwork):
    n_in : int #Size of the input vector
    n_cond : int #Size of the conditionning variable
    n_layers : int #Number of layers (i.e. number of stacked MADEs)
    layers : list[int] #list of hidden dimensionsin each MADE.
    activation : Callable #Activation function
    use_reverse : bool =True #Whether to reverse the order of the input between each MADE
    seed : Optional[int] = None #Random seed to label nodes
    

    def setup(self):
        np.random.seed(self.seed)
        layer_list = []
        for _ in range(self.n_layers):
            layer_list.append(
                MAFLayer(
                    n_in=self.n_in, n_cond=self.n_cond, hidden_dims=self.layers, reverse=self.use_reverse, seed=np.random.randint(0, 1000), activation=self.activation
                )
            )
        self.layer_list = layer_list
        self.mean = jnp.zeros(self.n_in)
        self.cov = jnp.eye(self.n_in)

    @nn.compact
    def __call__(self, x, y=None):
        """
        Forward pass of the model. Returns mean and variance of the gaussian conditionals
        as well as the log-determinant of the Jacobian

        Parameters
        ----------
        x : jnp.Array
            Input vector
        y : jnp.Array
            Conditionning variable
        """
        log_det_sum = jnp.zeros(x.shape[0])
        for layer in self.layer_list:
            x, log_det = layer.forward(x, y)
            log_det_sum += log_det
            #x = nn.BatchNorm(use_running_average=not train)(x)
        return x, log_det_sum
    
    def backward(self, u, y=None):
        log_det_sum = jnp.zeros(u.shape[0])
        #backward pass
        for layer in reversed(self.layer_list):
            u, log_det = layer.backward(u, y)
            log_det_sum += log_det
        return u, log_det_sum
    
    def log_prob(self, x, y=None):
        u, log_det_sum = self.__call__(x, y)
        log_pdf = multivariate_normal.logpdf(u, self.mean, self.cov)
        return log_pdf + log_det_sum
    
    def sample(self, y=None, num_samples=1, key=None):
        u = jax.random.multivariate_normal(key, self.mean, self.cov, shape=(num_samples,))
        if y is not None:
            y = y*jnp.ones((num_samples, 1))
        x, _ = self.backward(u, y)
        return x


        



from typing import Any, Callable, Dict, Optional, Union
from jaxtyping import Array, Float, PyTree
import jax
import jax.numpy as jnp
import jax.random as jr

import jaxili
from jaxili.utils import (
    check_density_estimator,
    validate_theta_x
)
from jaxili.train import TrainerModule
from jaxili.loss import loss_nll_npe

import jax_dataloader as jdl

class NPE():

    def __init__(
            self,
            model_class: jaxili.model.NDENetwork,
            model_hparams: Dict[str, Any],
            logging_level: Union[int, str] = "WARNING",
            show_progress_bar: bool = True
    ):
        """
        Base class for Neural Posterior Estimation (NPE) methods.

        Parameters
        ----------
        model_class : jaxili.model.NDENetwork
            Class of the neural density estimator to use.
        model_hparams : Dict[str, Any]
            Hyperparameters to use for the model.
        logging_level: Union[int, str], optional
            Logging level to use. Default is "WARNING".
        show_progress_bar : bool, optional
            Whether to show a progress bar during training. Default is True.
        """
        self._model_class = model_class
        self._model_hparams = model_hparams
        self._logging_level = logging_level
        self.show_progress_bar = show_progress_bar        

    
    def append_simulations(
            self,
            theta: Array,
            x: Array,        
    ):
        """
        Store parameters and simulation outputs to use them for later training.

        Data is stored in a Dataset object from `jax-dataloader`

        Parameter
        ---------
        theta : Array
            Parameters of the simulations.
        x : Array
            Simulation outputs.
        """
        #Verify theta and x typing and batch_size
        theta, x, batch_size = validate_theta_x(theta, x)

        self._dim_params = theta.shape[1]
        self._dim_outputs = x.shape[1]
        self._batch_size = batch_size

        #Create a Dataset
        self._dataset = jdl.ArrayDataset(theta, x)

        return self
    
    def create_trainer(self,
                       optimizer_hparams: Dict[str, Any],
                       exmp_input: Any,
                       seed: int = 42,
                       logger_params : Dict[str, Any] = None,
                       debug : bool = False,
                       check_val_every_epoch : int = 1):
        """
        Create a TrainerModule for the density estimator.

        Parameters
        ----------
        optimizer_hparams : Dict[str, Any]
            Hyperparameters to use for the optimizer.
        loss_fn : Callable
            Loss function to use for training.
        exmp_input : Any
            Example input to use for the model.
        seed : int, optional
            Seed to use for the trainer. Default is 42.
        logger_params : Dict[str, Any], optional
            Parameters to use for the logger. Default is None.
        debug : bool, optional
            Whether to use debug mode. Default is False.
        check_val_every_epoch : int, optional
            Frequency at which to check the validation loss. Default is 1.
        """

        self.trainer = TrainerModule(
            model_class = self._model_class,
            model_hparams = self._model_hparams,
            optimizer_hparams = optimizer_hparams,
            loss_fn = loss_nll_npe,

        )
    
    def train(
            self,
            training_batch_size: int = 50,
            learning_rate: float =  5e-4,
            validation_size: float = 0.1,
            patience: int = 20,
            n_epoch: int = 2**31 -1,
            model_hparams: Optional[Dict] = None,
            key: Optional[Array] = None,
            checkpoint_path: Optional[str] = '.'

    ):
        r"""
        Train the density estimator to approximate the distribution $p(\theta|x)$.

        Parameters
        ----------
        training_batch_size : int, optional
            Batch size to use during training. Default is 50.
        learning_rate: float, optional
            Learning rate to use during training. Default is 5e-4.
        validation_size: float, optional
            Fraction of the dataset to use for validation. Default is 0.1.
        patience: int, optional
            Number of epochs to wait before early stopping. Default is 20.
        n_epoch: int, optional
            Maximum number of epochs to train. Default is 2**31 - 1.
        model_hparams: Optional[Dict], optional
            Hyperparameters to use for the model. Default is None.
        key: Optional[Array], optional
            Random key to use for training. Default is None.
        """

        #Create a training and validation DataLoader using the dataset.
        assert (self._dataset is not None), "No dataset found. Please append simulations first."

        if key is None:
            key = jr.PRNGKey(0)
        key, subkey = jr.split(key)

        permutation = jr.permutation(subkey, self._batch_size)

        index_max = int((1 - validation_size) * self._batch_size)

        train_idx = permutation[:index_max]
        val_idx = permutation[index_max:]

        train_ds = jdl.ArrayDataset(*self._dataset[train_idx])
        val_ds = jdl.ArrayDataset(*self._dataset[val_idx])

        train_loader = jdl.DataLoader(
            train_ds,
            'jax',
            batch_size=training_batch_size,
            shuffle=True,
            drop_last=False
        )
        val_loader =  jdl.DataLoader(
            val_ds,
            'jax',
            batch_size=training_batch_size,
            shuffle=False,
            drop_last=False
        )

        #Build the trainer with correct architecture
        self.trainer = TrainerModule(
            
        )


